=head1 NAME

RemoteJob -- Subroutines needed to talk to the makepp job server

=head1 SYNOPSYS


   $remote_job = new RemoteJob(machine_name, ostype, environment);
                                # Returns a handle to talk to job on 
                                # remote machine.
   $remote_job = new RemoteJob; # Same, except that the remote machine is
                                # picked (somewhat arbitrarily).

   ($status, $stdout) = $remote_job->execute("command");
                                # Execute the command with the given output.
   $remote_job->close;          # Done with the job.

=head1 DESCRIPTION

RemoteJob is makepp's mechanism for handling parallel builds on a computation
farm.  It is essentially an extremely light-weight batch system which can
be set up by users without system manager assistance.

To use this, simply create a RemoteJob object using the new operator.
If you know which machine you want to execute it on, or what the architecture
type is, you can specify this as an argument to the constructor; otherwise
it will pick automatically.  A socket object is returned which you use to
communicate to the remote job.  The command will be executed on the remote
machine with its standard input/output/error all talking to the socket.

No environment variables or other setup has been done on the remote machine.
Thus you will probably not want to execute a program directly; you will
probably want to use it like this:

   $remote_job = new RemoteJob("/bin/sh");
   print $remote_job "cd $dir\n";
   while (($var, $val) = each %ENV) {
       print $remote_job "$var=$val\nexport $var\n";
   }
   print $remote_job "$command_to_execute\n";

RemoteJob::new returns an undefined object if some kind of error occured.

=cut

package RemoteJob;

use IO::Socket;
use Digest::MD5 qw(md5_base64);

@ISA = qw(IO::Socket);

$server_dirs = "$ENV{'HOME'}/.makepp_job_server";

sub new {
  my ($package, $machine, $ostype, $env) = @_; # Name the arguments.

  $env ||= \%ENV;
  $ostype ||= $^O;              # Pick the type if the user didn't.

  my $server;
  if (defined($machine)) {      # Known host?
    $server = start_remote_process($machine);
  } else {
#
# Pick a machine to use.  Currently we use a very unsophisticated algorithm
# here: we just keep trying machines in order.  We ought to do something
# better, but....
#
    die "This doesn't work yet when you don't specify the machine explicitly.\n";
  }

  return bless $server, $package;
}

#
# Start up a process on a remote server:
#
sub start_remote_process {
  my $machine = $_[0];

  local $_;
  unless (defined($server_info{$machine})) { # Don't already have info?
    my $server_info_dir = "$server_dirs/$ostype/$machine";
                                # Get the directory describing the machine.
    local *SERVER_INFO;
    open(SERVER_INFO, "$server_info_dir") or
      die "can't open $server_info_dir--$!\n";
    my ($server_key, $server_host, $server_port);
    while (defined($_ = <SERVER_INFO>)) {
      if (/^key (\w+)/) {
        $serer_info{$machine}{KEY} = $1;
      } elsif (/^host (\S+)/) {
        $server_info{$machine}{HOST} = $1;
      } elsif (/^port (\d+)/) {
        $server_info{$machine}{PORT} = $1;
      }
    }
    close SERVER_INFO;
  }


  my $server = new IO::Socket::INET(PeerAddr => $server_info{$machine}{HOST},
                                    PeerPort => $server_info{$machine}{PORT},
                                    Proto => "tcp",
                                    Type => SOCK_STREAM);
  $server or return $server;    # Some error occured.

  while (defined($_ = <$server>)) { # Get the authentication challenge.
    if (/^authenticate (\S+)/) {
      print $server md5_base64($server_info{$machine}{KEY} . $1), "\n";
    }
    elsif (/^ok/) {             # Ready for instructions?
      last;
    }
    elsif (/^too busy/) {
      $server->close;           # Can't do anything with this.
      return undef;
    }
  }
}

#
# Execute a command:
#
sub execute {
  my ($self, $command) = @_;    # Get the arguments.

  chomp $command;

  print $self "$command\n";     # Send the command.
  my $stdout = '';              # Where we accumulate the standard output.
  my $status = 1;               # Assume it will abort.
  local $_;
  while (defined($_ = <$self>)) { # Get the standard output:
    if (/^\05makepp_status=(\d+)$/) { # Status report?
      $status = $1;             # Set the status.
      last;                     # That's the end of this command.
    } else {
      $stdout .= $_;
    }
  }

  return ($status, $stdout);
}
